package ru.dbotthepony.mc.otm.client.screen

import com.mojang.blaze3d.vertex.PoseStack
import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectFunction
import net.minecraft.world.inventory.Slot
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.client.render.element
import ru.dbotthepony.mc.otm.client.screen.panels.*
import ru.dbotthepony.mc.otm.menu.ExoSuitInventoryMenu
import ru.dbotthepony.mc.otm.network.ExoSuitMenuOpen
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel

class ExoSuitInventoryScreen(menu: ExoSuitInventoryMenu) : MatteryScreen<ExoSuitInventoryMenu>(menu, TranslatableComponent("otm.gui.exosuit")) {
	override fun makeMainFrame(): FramePanel {
		val frame = FramePanel(this, width = 200f, height = 180f, title = this.title)

		val toolbeltLine = EditablePanel(this, frame, height = 18f)
		toolbeltLine.dock = Dock.BOTTOM

		toolbeltLine.setDockMargin(top = 3f)

		var mainInventoryLine: EditablePanel? = null

		val scrollPanel = DiscreteScrollBarPanel(this, null, maxScroll = { ((menu.playerCombinedInventorySlots.size - 27) + 8) / 9 },
			scrollCallback = {
					_, old, new ->

				for (i in old .. old + 2) {
					getInventorySlotsRow(i).visible = false
				}

				for (i in new .. new + 2) {
					val row = getInventorySlotsRow(i)
					row.visible = true
					row.y = (i - new) * AbstractSlotPanel.SIZE
					row.parent = mainInventoryLine
				}

				menu.lastScroll = new
			})

		scrollPanel.scroll = menu.lastScroll

		mainInventoryLine = object : EditablePanel(this@ExoSuitInventoryScreen, frame, height = 18f * 3f) {
			override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
				return scrollPanel.mouseScrolledInner(x, y, scroll)
			}
		}

		scrollPanel.parent = mainInventoryLine

		mainInventoryLine.dock = Dock.BOTTOM

		for (slot in menu.playerHotbarSlots) {
			val panel = SlotPanel(this, toolbeltLine, slot)
			panel.dock = Dock.LEFT
		}

		val offhand = SlotPanel(this, toolbeltLine, menu.offhandSlot)
		offhand.dock = Dock.RIGHT

		for (i in scrollPanel.scroll .. scrollPanel.scroll + 2) {
			getInventorySlotsRow(i).also {
				it.parent = mainInventoryLine
				it.y = AbstractSlotPanel.SIZE * (i - scrollPanel.scroll)
			}
		}

		val topLine = EditablePanel(this, frame, height = 18f * 4f)
		topLine.dock = Dock.TOP
		topLine.setDockMargin(top = 3f)

		val armorSlotsStrip = EditablePanel(this, topLine, width = 18f)
		armorSlotsStrip.dock = Dock.LEFT

		for (slot in menu.armorSlots) {
			val panel = SlotPanel(this, armorSlotsStrip, slot)
			panel.dock = Dock.TOP
		}

		val playerRectangle = EntityRendererPanel(this, topLine, minecraft!!.player!!)

		playerRectangle.dock = Dock.LEFT
		playerRectangle.setDockMargin(right = 3f)

		val craftingCanvas = EditablePanel(this, topLine, width = 90f)
		craftingCanvas.dock = Dock.LEFT

		val craftingSlotsCanvas = EditablePanel(this, craftingCanvas)

		if (menu.capability.isExoSuitCraftingUpgraded) {
			craftingSlotsCanvas.setSize(54f, 54f)

			for (row in 0 .. 2) {
				for (column in 0 .. 2) {
					SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[row * 3 + column], x = column * 18f, y = row * 18f)
				}
			}
		} else {
			craftingSlotsCanvas.setSize(36f, 36f)

			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[0])
			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[1], x = 18f)
			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[2], y = 18f)
			SlotPanel(this, craftingSlotsCanvas, menu.craftingSlots[3], x = 18f, y = 18f)
		}

		craftingSlotsCanvas.y = topLine.height / 2f - craftingSlotsCanvas.height / 2f

		val resultPanel = SlotPanel(this, craftingCanvas, menu.craftingResultSlot, x = craftingCanvas.width - 18f, y = topLine.height / 2f - 9f)

		val arrowPanel = object : EditablePanel(
			this@ExoSuitInventoryScreen,
			craftingCanvas,
			x = craftingSlotsCanvas.width,
			width = craftingCanvas.width - resultPanel.width - craftingSlotsCanvas.width,
			height = topLine.height
		) {
			override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
				CRAFT_ARROW.render(stack, x = width / 2f - CRAFT_ARROW.w / 2f, y = height / 2f - CRAFT_ARROW.h / 2f)
			}
		}

		scrollPanel.dock = Dock.RIGHT
		scrollPanel.setDockMargin(right = 3f)

		return frame
	}

	init {
		MatteryPlayerNetworkChannel.sendToServer(ExoSuitMenuOpen)
		ru.dbotthepony.mc.otm.client.minecraft.player?.containerMenu = menu
	}

	override fun onClose() {
		super.onClose()
	}

	companion object {
		val ENTITY_RECTANGLE = INVENTORY_LOCATION.element(25f, 7f, 51f, 72f)
		val STATUS_EFFECT_BG = INVENTORY_LOCATION.element(0f, 166f, 120f, 32f)
		val CRAFT_ARROW = INVENTORY_LOCATION.element(135f, 29f, 16f, 13f)
	}
}
