package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.ChatFormatting
import net.minecraft.client.gui.components.Button
import net.minecraft.client.gui.components.Button.OnPress
import net.minecraft.network.chat.Component
import ru.dbotthepony.mc.otm.client.playGuiClickSound
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.client.render.*
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.core.next
import ru.dbotthepony.mc.otm.core.prev
import java.util.*
import kotlin.collections.ArrayList
import kotlin.reflect.KMutableProperty0

open class ButtonPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 40f,
	height: Float = HEIGHT,
	label: Component
) : MinecraftWidgetPanel<Button>(screen, parent, x, y, width, height) {
	constructor(screen: MatteryScreen<*>, parent: EditablePanel?, label: Component) : this(screen, parent, x = 0f, label = label)
	constructor(
		screen: MatteryScreen<*>,
		parent: EditablePanel?,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = 40f,
		height: Float = HEIGHT,
		label: Component,
		onPress: Runnable
	) : this(screen, parent, x, y, width, height, label) {
		this.callback = onPress
	}

	var label = label
		set(value) {
			field = value
			widget?.message = value
		}

	override fun copyValues(new_widget: Button, old_widget: Button) {}
	override fun configureNew(widget: Button, recreation: Boolean) {}

	override fun makeNew(): Button {
		return object : Button(0, 0, width.toInt(), height.toInt().coerceAtMost(20), label, OnPress { this@ButtonPanel.onPress() }) {
			override fun isHoveredOrFocused(): Boolean {
				return this@ButtonPanel.isHovered
			}
		}
	}

	protected var callback: Runnable? = null

	fun bind(runnable: Runnable) {
		callback = runnable
	}

	protected open fun onPress() {
		callback?.run()
	}

	companion object {
		const val HEIGHT = 20f
	}
}

@Suppress("PropertyName")
abstract class SquareButtonPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float,
	height: Float,
	val skinElement: (() -> SkinElement)? = null,
	val lambdaOnPress: ((clickButton: Int) -> Unit)? = null,
) : EditablePanel(screen, parent, x, y, width, height) {
	protected var pressed = false

	protected open fun click(clickButton: Int) {
		lambdaOnPress?.invoke(clickButton)
	}

	abstract val PRESSED: SkinElement
	abstract val HOVERED: SkinElement
	abstract val IDLE: SkinElement

	var isDisabled = false
		set(value) {
			if (field != value) {
				if (!value) {
					pressed = false
					trapMouseInput = false
				}

				field = value
			}
		}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		if (pressed) {
			PRESSED.render(stack, 0f, 0f, width, height)
		} else if (isHovered && !isDisabled) {
			HOVERED.render(stack, 0f, 0f, width, height)
		} else {
			IDLE.render(stack, 0f, 0f, width, height)
		}

		skinElement?.invoke()?.render(stack, 0f, 0f, width, height)
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		if (!isDisabled) {
			if (!pressed) {
				playGuiClickSound()
			}

			pressed = true
			trapMouseInput = true
		}

		return true
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		if (!isDisabled && pressed) {
			pressed = false

			if (isHovered) {
				click(button)
			}
		}

		trapMouseInput = false

		return true
	}
}

abstract class EnumSquareButtonPanel<T : Enum<T>>(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float,
	height: Float,
	val enum: Class<T>,
	val prop: KMutableProperty0<T>,
	val defaultValue: T,
	val lambdaOnChange: ((newValue: T) -> Unit)? = null,
) : SquareButtonPanel(screen, parent, x, y, width, height, null, null) {
	private var building = true

	protected val enumMapping = EnumMap<T, Pair<SkinElement, UVWindingOrder>>(enum)
	protected val tooltipMapping = EnumMap<T, Component>(enum)

	fun addTooltip(value: T, component: Component): EnumSquareButtonPanel<T> {
		check(tooltipMapping.put(value, component) == null) { "Already has mapping for $value" }
		return this
	}

	var mainTooltip: Component? = null
		set(value) {
			if (field != null) {
				throw UnsupportedOperationException("Write once only")
			}

			field = value
		}

	fun isFullyDefined(): Boolean {
		for (value in enum.enumConstants) {
			if (enumMapping[value] == null) {
				return false
			}
		}

		return true
	}

	val missingValues: List<T> get() {
		val missing = ArrayList<T>()

		for (value in enum.enumConstants) {
			if (enumMapping[value] == null) {
				missing.add(value)
			}
		}

		return missing
	}

	fun add(value: T, skinElement: SkinElement, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumSquareButtonPanel<T> {
		return add(value, skinElement to winding)
	}

	fun add(value: T, skinElement: SkinElement, component: Component, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumSquareButtonPanel<T> {
		return add(value, component, skinElement to winding)
	}

	fun add(value: T, component: Component, skinElement: SkinElement, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumSquareButtonPanel<T> {
		return add(value, component, skinElement to winding)
	}

	fun add(value: T, pair: Pair<SkinElement, UVWindingOrder>): EnumSquareButtonPanel<T> {
		check(building) { "Not building" }
		check(enumMapping.put(value, pair) == null) { "Already has mapping for $value" }

		if (enumMapping.size == enum.enumConstants.size) {
			finish()
		}

		return this
	}

	fun add(value: T, component: Component, pair: Pair<SkinElement, UVWindingOrder>): EnumSquareButtonPanel<T> {
		addTooltip(value, component)
		return add(value, pair)
	}

	fun finish(): EnumSquareButtonPanel<T> {
		check(building) { "Not building" }
		check(isFullyDefined()) {
			"Not all enums having their mapping defined, missing are: ${missingValues.joinToString(", ")}"
		}

		building = false
		return this
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		check(!building) { "Still building this button!" }
		super.innerRender(stack, mouseX, mouseY, partialTick)
		val pair = checkNotNull(enumMapping[prop.get()]) { "HOW" }
		pair.first.render(stack, 0f, 0f, width, height, pair.second)
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		check(!building) { "Still building this button!" }
		return super.mouseClickedInner(x, y, button)
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		check(!building) { "Still building this button!" }
		return super.mouseReleasedInner(x, y, button)
	}

	override fun click(clickButton: Int) {
		when (clickButton) {
			InputConstants.MOUSE_BUTTON_LEFT -> {
				prop.set(prop.get().next(enum.enumConstants))
				lambdaOnChange?.invoke(prop.get())
			}

			InputConstants.MOUSE_BUTTON_RIGHT -> {
				prop.set(prop.get().prev(enum.enumConstants))
				lambdaOnChange?.invoke(prop.get())
			}

			InputConstants.MOUSE_BUTTON_MIDDLE -> {
				if (prop.get() != defaultValue) {
					prop.set(defaultValue)
					lambdaOnChange?.invoke(prop.get())
				}
			}
		}
	}

	override fun innerRenderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		if (!isHovered && !isTrappingMouseInput()) {
			return super.innerRenderTooltips(stack, mouse_x, mouse_y, flag)
		}

		if (mainTooltip == null && tooltipMapping.size == 0) {
			return super.innerRenderTooltips(stack, mouse_x, mouse_y, flag)
		}

		val listing = ArrayList<Component>()

		if (mainTooltip != null) {
			listing.add(mainTooltip!!)
			listing.add(SPACE)
		}

		for ((key, value) in tooltipMapping) {
			if (key == prop.get()) {
				listing.add(value.copy().withStyle(ChatFormatting.WHITE))
			} else {
				listing.add(value.copy().withStyle(ChatFormatting.GRAY))
			}
		}

		screen.renderComponentTooltip(
			stack,
			listing,
			mouse_x.toInt(),
			mouse_y.toInt(),
			font
		)

		return true
	}

	companion object {
		private val SPACE = TextComponent("")
	}
}

open class LargeSquareButtonPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = SIZE,
	height: Float = SIZE,
	skinElement: (() -> SkinElement)? = null,
	lambdaOnPress: ((clickButton: Int) -> Unit)? = null,
) : SquareButtonPanel(screen, parent, x, y, width, height, skinElement, lambdaOnPress) {
	final override val IDLE = Widgets18.BUTTON_IDLE
	final override val HOVERED = Widgets18.BUTTON_HOVERED
	final override val PRESSED = Widgets18.BUTTON_PRESSED

	companion object {
		const val SIZE = 18f
	}
}
open class LargeEnumSquareButtonPanel<T : Enum<T>>(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = SIZE,
	height: Float = SIZE,
	enum: Class<T>,
	prop: KMutableProperty0<T>,
	defaultValue: T,
	lambdaOnChange: ((newValue: T) -> Unit)? = null,
) : EnumSquareButtonPanel<T>(screen, parent, x, y, width, height, enum, prop, defaultValue, lambdaOnChange) {
	final override val IDLE = Widgets18.BUTTON_IDLE
	final override val HOVERED = Widgets18.BUTTON_HOVERED
	final override val PRESSED = Widgets18.BUTTON_PRESSED

	companion object {
		const val SIZE = 18f
	}
}

open class SmallSquareButtonPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = SIZE,
	height: Float = SIZE,
	skinElement: (() -> SkinElement)? = null,
	lambdaOnPress: ((clickButton: Int) -> Unit)? = null,
) : SquareButtonPanel(screen, parent, x, y, width, height, skinElement, lambdaOnPress) {
	final override val IDLE = Widgets8.BUTTON_IDLE
	final override val HOVERED = Widgets8.BUTTON_HOVERED
	final override val PRESSED = Widgets8.BUTTON_PRESSED

	companion object {
		const val SIZE = 8f
	}
}

open class SmallEnumSquareButtonPanel<T : Enum<T>>(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = SIZE,
	height: Float = SIZE,
	enum: Class<T>,
	prop: KMutableProperty0<T>,
	defaultValue: T,
	lambdaOnChange: ((newValue: T) -> Unit)? = null,
) : EnumSquareButtonPanel<T>(screen, parent, x, y, width, height, enum, prop, defaultValue, lambdaOnChange) {
	final override val IDLE = Widgets8.BUTTON_IDLE
	final override val HOVERED = Widgets8.BUTTON_HOVERED
	final override val PRESSED = Widgets8.BUTTON_PRESSED

	companion object {
		const val SIZE = 8f
	}
}
